/*
* Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
*
* This software product is a proprietary product of Nvidia Corporation and its affiliates
* (the "Company") and all right, title, and interest in and to the software
* product, including all associated intellectual property rights, are and
* shall remain exclusively with the Company.
*
* This software product is governed by the End User License Agreement
* provided with the software product.
*/

#include <netlink/route/lwt/api.h>
#include <endian.h>
#include <errno.h>
#include <inttypes.h>
#include <linux/lwtunnel.h>
#include <netlink-private/netlink.h>
#include <netlink/attr.h>
#include <netlink/route/lwt/ip.h>

static struct nla_policy lwt_ip_policy[LWTUNNEL_IP_MAX + 1] = {
	[LWTUNNEL_IP_ID]	= { .type = NLA_U64 },
	[LWTUNNEL_IP_DST]	= { .type = NLA_U32 },
	[LWTUNNEL_IP_SRC]	= { .type = NLA_U32 },
	[LWTUNNEL_IP_TTL]	= { .type = NLA_U8 },
	[LWTUNNEL_IP_TOS]	= { .type = NLA_U8 },
	[LWTUNNEL_IP_FLAGS]	= { .type = NLA_U16 },
};

struct rtnl_lwt_ip_encap *rtnl_lwt_ip_encap_hdr(struct rtnl_lwt_encap *encap)
{
	return (struct rtnl_lwt_ip_encap *)encap->l_hdr;
}

static int lwt_ip_build_encap(struct nlattr *nla,
			      struct rtnl_lwt_encap **encap)
{
	struct nlattr *tb[LWTUNNEL_IP_MAX + 1];
	struct rtnl_lwt_encap *newencap;
	struct rtnl_lwt_ip_encap *ip_encap;
	int err;

	err = nla_parse_nested(tb, LWTUNNEL_IP_MAX, nla, lwt_ip_policy);
	if (err < 0)
		goto out_err;

	newencap = malloc(sizeof(struct rtnl_lwt_encap) +
			  sizeof(struct rtnl_lwt_ip_encap));
	if (!newencap) {
		err = -ENOMEM;
		goto out_err;
	}

	newencap->l_type = LWTUNNEL_ENCAP_IP;
	newencap->l_len = sizeof(struct rtnl_lwt_ip_encap);

	ip_encap = rtnl_lwt_ip_encap_hdr(newencap);
	ip_encap->li_id = be64toh(nla_get_u64(tb[LWTUNNEL_IP_ID]));
	ip_encap->li_dst = nl_addr_alloc_attr(tb[LWTUNNEL_IP_DST], AF_INET);
	if (!ip_encap->li_dst) {
		free(newencap);
		err = -ENOMEM;
		goto out_err;
	}

	*encap = newencap;
	return 0;

out_err:
	*encap = NULL;
	return err;
}

static struct rtnl_lwt_encap *lwt_ip_clone_encap(struct rtnl_lwt_encap *src)
{
	struct rtnl_lwt_ip_encap *src_encap, *dst_encap;
	struct rtnl_lwt_encap *newencap;

	newencap = malloc(sizeof(struct rtnl_lwt_encap) +
			  sizeof(struct rtnl_lwt_ip_encap));
	if (!newencap)
		return NULL;

	newencap->l_type = LWTUNNEL_ENCAP_IP;
	newencap->l_len = sizeof(struct rtnl_lwt_ip_encap);

	src_encap = rtnl_lwt_ip_encap_hdr(src);
	dst_encap = rtnl_lwt_ip_encap_hdr(newencap);
	dst_encap->li_id = src_encap->li_id;
	dst_encap->li_dst = nl_addr_clone(src_encap->li_dst);
	if (!dst_encap->li_dst) {
		free(newencap);
		return NULL;
	}

	return newencap;
}

static void lwt_ip_release_encap(struct rtnl_lwt_encap *encap)
{
	struct rtnl_lwt_ip_encap *ip_encap;

	if (!encap)
		return;

	ip_encap = rtnl_lwt_ip_encap_hdr(encap);
	nl_addr_put(ip_encap->li_dst);
	free(encap);
}

static int lwt_ip_compare_encap(struct rtnl_lwt_encap *a,
				struct rtnl_lwt_encap *b)
{
	struct rtnl_lwt_ip_encap *ip_encap_a = rtnl_lwt_ip_encap_hdr(a);
	struct rtnl_lwt_ip_encap *ip_encap_b = rtnl_lwt_ip_encap_hdr(b);

	if (a->l_type != b->l_type || a->l_len != b->l_len)
		return 1;

	if (ip_encap_a->li_id != ip_encap_b->li_id)
		return 1;

	return nl_addr_cmp(ip_encap_a->li_dst, ip_encap_b->li_dst);
}

static void lwt_ip_dump_encap(struct rtnl_lwt_encap *encap,
			      struct nl_dump_params *p)
{
	const struct rtnl_lwt_ip_encap *ip_encap = rtnl_lwt_ip_encap_hdr(encap);
	char buf[64];

	nl_dump(p, "encap ip id %" PRIu64 " dst %s ", ip_encap->li_id,
		nl_addr2str(ip_encap->li_dst, buf, sizeof(buf)));
}

int rtnl_lwt_is_ip(struct rtnl_lwt_encap *encap)
{
	return encap->l_type == LWTUNNEL_ENCAP_IP;
}

static struct rtnl_lwt_ops lwt_ip_ops = {
	.lwt_type		= LWTUNNEL_ENCAP_IP,
	.lwt_build_encap	= &lwt_ip_build_encap,
	.lwt_clone_encap	= &lwt_ip_clone_encap,
	.lwt_release_encap	= &lwt_ip_release_encap,
	.lwt_compare_encap	= &lwt_ip_compare_encap,
	.lwt_dump_encap		= &lwt_ip_dump_encap,
};

static void __init lwt_ip_init(void)
{
	rtnl_lwt_ops_register(&lwt_ip_ops);
}

static void __exit lwt_ip_exit(void)
{
	rtnl_lwt_ops_unregister(&lwt_ip_ops);
}
