/*
* Copyright © 2021 NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
*
* This software product is a proprietary product of Nvidia Corporation and its affiliates
* (the "Company") and all right, title, and interest in and to the software
* product, including all associated intellectual property rights, are and
* shall remain exclusively with the Company.
*
* This software product is governed by the End User License Agreement
* provided with the software product.
*/

#include <errno.h>
#include <inttypes.h>
#include <stdarg.h>
#include <stdbool.h>

#include <netlink/route/link.h>
#include <netlink/route/lwt/ip.h>
#include <netlink/route/lwt/ip6.h>
#include <netlink/route/route.h>
#include <netlink/object.h>


static void _vpferrno(int errnum, bool do_errnum, const char* format, va_list ap)
{
	fprintf(stderr, "Error: ");
	vfprintf(stderr, format, ap);
	if (do_errnum) {
		char buf[256];
		fprintf(stderr, ": %s", strerror_r(errnum, buf, sizeof(buf)));
	}
	fprintf(stderr, "\n");

	exit(1);
}

static void pferrno(int errnum, const char* format, ...)
{
	va_list ap;

	va_start(ap, format);
	_vpferrno(errnum, true, format, ap);
	va_end(ap);
}

static void pferror(const char* format, ...)
{
	va_list ap;

	va_start(ap, format);
	_vpferrno(0, false, format, ap);
	va_end(ap);
}

static void check_nh(struct rtnl_nexthop *nh, void *arg)
{
	bool *has_encap = (bool *)arg;
	struct rtnl_lwt_encap *encap;

	encap = rtnl_route_nh_get_lwt_encap(nh);
	if (encap && (rtnl_lwt_is_ip(encap) || rtnl_lwt_is_ip6(encap))) {
		*has_encap = true;
	}
}

static void process_nh(struct rtnl_nexthop *nh, void *arg)
{
	struct rtnl_lwt_encap *encap;
	char buf[64];

	encap = rtnl_route_nh_get_lwt_encap(nh);
	if (!encap) {
		return;
	} else if (rtnl_lwt_is_ip(encap)) {
		struct rtnl_lwt_ip_encap *ip_encap;

		ip_encap = rtnl_lwt_ip_encap_hdr(encap);
		printf("\tencap ip id %" PRIu64 " dst %s\n", ip_encap->li_id,
		       nl_addr2str(ip_encap->li_dst, buf, sizeof(buf)));
	} else if (rtnl_lwt_is_ip6(encap)) {
		struct rtnl_lwt_ip6_encap *ip6_encap;

		ip6_encap = rtnl_lwt_ip6_encap_hdr(encap);
		nl_addr2str(ip6_encap->li6_dst, buf, sizeof(buf));
		printf("\tencap ip6 id %" PRIu64 " dst %s\n",
		       ip6_encap->li6_id, buf);
	}
}

static void process_route(struct nl_object *route_obj, void *arg)
{
	struct rtnl_route *route = (struct rtnl_route *)route_obj;
	struct nl_dump_params params = {
		.dp_fd = stdout,
		.dp_type = NL_DUMP_LINE,
	};
	bool has_encap = false;

	rtnl_route_foreach_nexthop(route, check_nh, &has_encap);

	if (has_encap) {
		nl_object_dump(route_obj, &params);
		rtnl_route_foreach_nexthop(route, process_nh, NULL);
	}
}

int main(int argc, char *argv[])
{
	struct nl_cache *link_cache, *route_cache;
	struct nl_sock *sock;
	int err;

	if (argc != 1) {
		fprintf(stderr, "Usage: %s\n", argv[0]);
		exit(2);
	}

	if (!(sock = nl_socket_alloc())) {
		pferrno(ENOBUFS, "Unable to allocate netlink socket");
	}
	if ((err = nl_connect(sock, NETLINK_ROUTE)) < 0) {
		pferror("Unable to connect netlink socket: %s",
			nl_geterror(err));
	}

	if ((err = rtnl_link_alloc_cache(sock, AF_UNSPEC, &link_cache)) < 0) {
		pferror("Unable to allocate link cache: %s",
			nl_geterror(err));
	}
	nl_cache_mngt_provide(link_cache);

	if ((err = rtnl_route_alloc_cache(sock, AF_UNSPEC, 0, &route_cache)) <
	    0) {
		pferror("Unable to allocate route cache: %s",
			nl_geterror(err));
	}
	nl_cache_mngt_provide(route_cache);

	nl_cache_foreach(route_cache, &process_route, NULL);

	nl_cache_mngt_unprovide(route_cache);
	nl_cache_free(route_cache);
	nl_cache_mngt_unprovide(link_cache);
	nl_cache_free(link_cache);
	nl_socket_free(sock);

	return 0;
}
